@@ -13,14 +13,14 @@ from pathlib import Path
 import pmb.build
 import pmb.chroot
 import pmb.chroot.apk
-import pmb.chroot.other
 import pmb.chroot.initfs
 import pmb.config
 import pmb.config.pmaports
 from pmb.helpers.locale import get_xkb_layout
+import pmb.parse.depends
 from pmb.parse.deviceinfo import Deviceinfo
 from pmb.core import Config
-from pmb.types import Env, PartitionLayout, PmbArgs
+from pmb.types import Env, PartitionLayout, PmbArgs, RunOutputTypeDefault
 import pmb.helpers.devices
 from pmb.helpers.mount import mount_device_rootfs
 import pmb.helpers.run
@@ -111,7 +111,7 @@ def get_kernel_package(config: Config) -> list[str]:
     return ["device-" + config.device + "-kernel-" + config.kernel]
 
 
-def copy_files_from_chroot(args: PmbArgs, chroot: Chroot) -> None:
+def copy_files_from_chroot(chroot: Chroot, rsync: bool, verbose: bool) -> None:
     """
     Copy all files from the rootfs chroot to /mnt/install, except
     for the home folder (because /home will contain some empty
@@ -125,7 +125,7 @@ def copy_files_from_chroot(args: PmbArgs, chroot: Chroot) -> None:
     mountpoint_outside = Chroot.native() / mountpoint
 
     # Remove empty qemu-user binary stub (where the binary was bind-mounted)
-    arch_qemu = pmb.parse.deviceinfo().arch.qemu()
+    arch_qemu = pmb.parse.deviceinfo().arch.qemu_user()
     qemu_binary = mountpoint_outside / f"usr/bin/qemu-{arch_qemu}-static"
     if os.path.exists(qemu_binary):
         pmb.helpers.run.root(["rm", qemu_binary])
@@ -143,10 +143,10 @@ def copy_files_from_chroot(args: PmbArgs, chroot: Chroot) -> None:
         folders.append(path.name)
 
     # Update or copy all files
-    if args.rsync:
+    if rsync:
         pmb.chroot.apk.install(["rsync"], Chroot.native())
         rsync_flags = "-a"
-        if args.verbose:
+        if verbose:
             rsync_flags += "vP"
         pmb.chroot.root(
             ["rsync", rsync_flags, "--delete", *folders, "/mnt/install/"], working_dir=mountpoint
@@ -177,7 +177,7 @@ def create_home_from_skel(filesystem: str, user: str) -> None:
     pmb.helpers.run.root(["chown", "-R", "10000", home])
 
 
-def configure_apk(args: PmbArgs) -> None:
+def configure_apk(install_local_pkgs: bool) -> None:
     """
     Copy over all official keys, and the keys used to compile local packages
     (unless --no-local-pkgs is set). Then copy the corresponding APKINDEX files
@@ -187,7 +187,7 @@ def configure_apk(args: PmbArgs) -> None:
     keys_dir = pmb.config.apk_keys_path
 
     # Official keys + local keys
-    if args.install_local_pkgs:
+    if install_local_pkgs:
         keys_dir = get_context().config.work / "config_apk_keys"
 
     # Copy over keys
@@ -286,7 +286,9 @@ def setup_login(args: PmbArgs, config: Config, chroot: Chroot) -> None:
         else:
             while True:
                 try:
-                    pmb.chroot.root(["passwd", config.user], chroot, output="interactive")
+                    pmb.chroot.root(
+                        ["passwd", config.user], chroot, output=RunOutputTypeDefault.INTERACTIVE
+                    )
                     break
                 except RuntimeError:
                     logging.info("WARNING: Failed to set the password. Try it one more time.")
@@ -308,7 +310,13 @@ def copy_ssh_keys(config: Config) -> None:
     keys = []
     for key in glob.glob(os.path.expanduser(config.ssh_key_glob)):
         with open(key) as infile:
-            keys += infile.readlines()
+            try:
+                keys += infile.readlines()
+            except UnicodeDecodeError:
+                logging.info(
+                    f"WARNING: {key} is not a valid SSH key. "
+                    "This file will not be copied to device."
+                )
 
     if not len(keys):
         logging.info(
@@ -319,10 +327,9 @@ def copy_ssh_keys(config: Config) -> None:
         return
 
     authorized_keys = Chroot.native() / "tmp/authorized_keys"
-    outfile = open(authorized_keys, "w")
-    for key in keys:
-        outfile.write(f"{key}")
-    outfile.close()
+    with open(authorized_keys, "w") as outfile:
+        for key in keys:
+            outfile.write(f"{key}")
 
     target = Chroot.native() / "mnt/install/home/" / config.user / ".ssh"
     pmb.helpers.run.root(["mkdir", target])
@@ -344,7 +351,16 @@ def setup_keymap(config: Config) -> None:
     options = deviceinfo.keymaps.split(" ")
     if config.keymap != "" and config.keymap is not None and config.keymap in options:
         layout, variant = config.keymap.split("/")
-        pmb.chroot.root(["setup-keymap", layout, variant], chroot, output="interactive")
+        if pmb.config.is_systemd_selected():
+            pmb.chroot.root(
+                ["systemd-firstboot", "--force", f"--keymap={variant}"],
+                chroot,
+                output=RunOutputTypeDefault.LOG,
+            )
+        else:
+            pmb.chroot.root(
+                ["setup-keymap", layout, variant], chroot, output=RunOutputTypeDefault.INTERACTIVE
+            )
 
         # Check xorg config
         xconfig = None
@@ -399,9 +415,11 @@ def setup_locale(chroot: Chroot, locale: str) -> None:
     # 10locale-pmos.sh gets sourced before 20locale.sh from
     # alpine-baselayout by /etc/profile. Since they don't override the
     # locale if it exists, it warranties we have preference
-    line = f"export LANG=${{LANG:-{shlex.quote(locale)}}}"
+    pmb.chroot.root(["sh", "-c", f"echo LANG={shlex.quote(locale)} > /etc/locale.conf"], chroot)
+    # musl-locales doesn't read from /etc/locale.conf, only from environment variables
+    # TODO: once musl implements locales (hopefully with /etc/locale.conf support) the following should be removed
     pmb.chroot.root(
-        ["sh", "-c", f"echo {shlex.quote(line)} > /etc/profile.d/10locale-pmos.sh"], chroot
+        ["sh", "-c", "echo source /etc/locale.conf > /etc/profile.d/10locale-pmos.sh"], chroot
     )
     # add keyboard layout related to locale and layout switcher
     xkb_layout = get_xkb_layout(locale)
@@ -599,9 +617,9 @@ def generate_binary_list(args: PmbArgs, chroot: Chroot, step: int) -> list[tuple
     binaries = (pmb.parse.deviceinfo().sd_embed_firmware or "").split(",")
 
     for binary_offset in binaries:
-        binary, _offset = binary_offset.split(":")
+        binary, offset_ = binary_offset.split(":")
         try:
-            offset = int(_offset)
+            offset = int(offset_)
         except ValueError:
             raise RuntimeError(f"Value for firmware binary offset is not valid: {offset}")
         binary_path = chroot / "usr/share" / binary
@@ -734,15 +752,18 @@ def sanity_check_disk_size(args: PmbArgs) -> None:
     human = f"{size / 2 / 1024 / 1024:.2f} GiB"
 
     # Warn if the size is larger than 100GiB
-    if not args.assume_yes and size > (100 * 2 * 1024 * 1024):
-        if not pmb.helpers.cli.confirm(
+    if (
+        not args.assume_yes
+        and size > (100 * 2 * 1024 * 1024)
+        and not pmb.helpers.cli.confirm(
             f"WARNING: The target disk ({devpath}) "
             "is larger than a usual SD card "
             "(>100GiB). Are you sure you want to "
             f"overwrite this {human} disk?",
             no_assumptions=True,
-        ):
-            raise RuntimeError("Aborted.")
+        )
+    ):
+        raise RuntimeError("Aborted.")
 
 
 def get_ondev_pkgver(args: PmbArgs) -> str:
@@ -763,33 +784,41 @@ def sanity_check_ondev_version(args: PmbArgs) -> None:
         )
 
 
-def get_partition_layout(reserve: bool | int, kernel: bool) -> PartitionLayout:
+def get_partition_layout(reserve: bool | int, kernel: bool, prep: bool | None) -> PartitionLayout:
     """
     :param reserve: create an empty partition between root and boot (pma#463)
     :param kernel: create a separate kernel partition before all other
                    partitions, e.g. for the ChromeOS devices with cgpt
+    :param prep: create a PReP Boot partition before root and no boot partition,
+                 e.g. for PowerVM or OpenFirmware POWER machines
     :returns: the partition layout, e.g. without reserve and kernel:
               {"kernel": None, "boot": 1, "reserve": None, "root": 2}
     """
     ret: PartitionLayout = {
         "kernel": None,
+        "prep": None,
         "boot": 1,
         "reserve": None,
         "root": 2,
     }
 
-    if kernel:
-        ret["kernel"] = 1
-        ret["boot"] += 1
-        ret["root"] += 1
+    if prep:
+        ret["prep"] = 1
+        ret["boot"] = None
+    else:
+        if kernel:
+            ret["kernel"] = 1
+            ret["boot"] = 2
+            ret["root"] += 1
+
+        if reserve:
+            ret["reserve"] = ret["root"]
+            ret["root"] += 1
 
-    if reserve:
-        ret["reserve"] = ret["root"]
-        ret["root"] += 1
     return ret
 
 
-def get_uuid(args: PmbArgs, partition: Path) -> str:
+def get_uuid(partition: Path) -> str:
     """
     Get UUID of a partition
 
@@ -808,27 +837,27 @@ def get_uuid(args: PmbArgs, partition: Path) -> str:
     ).rstrip()
 
 
-def create_crypttab(args: PmbArgs, layout: PartitionLayout, chroot: Chroot) -> None:
+def create_crypttab(layout: PartitionLayout | None, chroot: Chroot) -> None:
     """
     Create /etc/crypttab config
 
-    :param layout: partition layout from get_partition_layout()
+    :param layout: partition layout from get_partition_layout() or None
     :param suffix: of the chroot, which crypttab will be created to
     """
+    root_dev = Path(f"/dev/installp{layout['root']}") if layout else Path("/dev/install")
 
-    luks_uuid = get_uuid(args, Path("/dev") / f"installp{layout['root']}")
-
+    luks_uuid = get_uuid(root_dev)
     crypttab = f"root UUID={luks_uuid} none luks\n"
 
     (chroot / "tmp/crypttab").open("w").write(crypttab)
     pmb.chroot.root(["mv", "/tmp/crypttab", "/etc/crypttab"], chroot)
 
 
-def create_fstab(args: PmbArgs, layout: PartitionLayout, chroot: Chroot) -> None:
+def create_fstab(args: PmbArgs, layout: PartitionLayout | None, chroot: Chroot) -> None:
     """
     Create /etc/fstab config
 
-    :param layout: partition layout from get_partition_layout()
+    :param layout: partition layout from get_partition_layout() or None
     :param chroot: of the chroot, which fstab will be created to
     """
 
@@ -837,20 +866,22 @@ def create_fstab(args: PmbArgs, layout: PartitionLayout, chroot: Chroot) -> None
     if args.on_device_installer and chroot.type == ChrootType.ROOTFS:
         return
 
-    boot_dev = Path(f"/dev/installp{layout['boot']}")
-    root_dev = Path(f"/dev/installp{layout['root']}")
+    uses_prep = layout and layout["prep"] is not None
+
+    if layout and not uses_prep:
+        boot_dev = Path(f"/dev/installp{layout['boot']}")
+        root_dev = Path(f"/dev/installp{layout['root']}")
+    elif layout and uses_prep:
+        boot_dev = None
+        root_dev = Path(f"/dev/installp{layout['root']}")
+    else:
+        boot_dev = None
+        root_dev = Path("/dev/install")
 
-    boot_mount_point = f"UUID={get_uuid(args, boot_dev)}"
     root_mount_point = (
-        "/dev/mapper/root" if args.full_disk_encryption else f"UUID={get_uuid(args, root_dev)}"
+        "/dev/mapper/root" if args.full_disk_encryption else f"UUID={get_uuid(root_dev)}"
     )
-
-    boot_options = "nodev,nosuid,noexec"
-    boot_filesystem = pmb.parse.deviceinfo().boot_filesystem or "ext2"
-    if boot_filesystem in ("fat16", "fat32"):
-        boot_filesystem = "vfat"
-        boot_options += ",umask=0077,nosymfollow,codepage=437,iocharset=ascii"
-    root_filesystem = pmb.install.get_root_filesystem(args)
+    root_filesystem = pmb.install.get_root_filesystem(args.filesystem)
 
     if root_filesystem == "btrfs":
         # btrfs gets separate subvolumes for root, var and home
@@ -862,17 +893,23 @@ def create_fstab(args: PmbArgs, layout: PartitionLayout, chroot: Chroot) -> None
 {root_mount_point} /srv btrfs subvol=@srv,compress=zstd:2,ssd 0 0
 {root_mount_point} /var btrfs subvol=@var,ssd 0 0
 {root_mount_point} /.snapshots btrfs subvol=@snapshots,compress=zstd:2,ssd 0 0
-
-{boot_mount_point} /boot {boot_filesystem} {boot_options} 0 0
 """.lstrip()
 
     else:
         fstab = f"""
 # <file system> <mount point> <type> <options> <dump> <pass>
 {root_mount_point} / {root_filesystem} defaults 0 0
-{boot_mount_point} /boot {boot_filesystem} {boot_options} 0 0
 """.lstrip()
 
+    if boot_dev:
+        boot_mount_point = f"UUID={get_uuid(boot_dev)}"
+        boot_options = "nodev,nosuid,noexec"
+        boot_filesystem = pmb.parse.deviceinfo().boot_filesystem or "ext2"
+        if boot_filesystem in ("fat16", "fat32"):
+            boot_filesystem = "vfat"
+            boot_options += ",umask=0077,nosymfollow,codepage=437,iocharset=ascii"
+        fstab += f"{boot_mount_point} /boot {boot_filesystem} {boot_options} 0 0\n"
+
     with (chroot / "tmp/fstab").open("w") as f:
         f.write(fstab)
     pmb.chroot.root(["mv", "/tmp/fstab", "/etc/fstab"], chroot)
@@ -887,6 +924,7 @@ def install_system_image(
     boot_label: str = "pmOS_boot",
     root_label: str = "pmOS_root",
     split: bool = False,
+    single_partition: bool = False,
     disk: Path | None = None,
 ) -> None:
     """
@@ -906,21 +944,31 @@ def install_system_image(
     logging.info(f"*** ({step}/{steps}) PREPARE INSTALL BLOCKDEVICE ***")
     pmb.helpers.mount.umount_all(chroot.path)
     (size_boot, size_root) = get_subpartitions_size(chroot)
-    layout = get_partition_layout(
-        size_reserve, bool(pmb.parse.deviceinfo().cgpt_kpart and args.install_cgpt)
-    )
+    if not single_partition:
+        layout = get_partition_layout(
+            size_reserve,
+            bool(pmb.parse.deviceinfo().cgpt_kpart) and args.install_cgpt,
+            pmb.parse.deviceinfo().create_prep_boot,
+        )
+    else:
+        layout = None
     if not args.rsync:
-        pmb.install.blockdevice.create(args, size_boot, size_root, size_reserve, split, disk)
-        if not split:
+        pmb.install.blockdevice.create(
+            args.sector_size, size_boot, size_root, size_reserve, split, disk
+        )
+        if not split and layout:
             if pmb.parse.deviceinfo().cgpt_kpart and args.install_cgpt:
                 pmb.install.partition_cgpt(layout, size_boot, size_reserve)
+            elif pmb.parse.deviceinfo().create_prep_boot:
+                pmb.install.partition_prep(layout)
             else:
                 pmb.install.partition(layout, size_boot, size_reserve)
 
     # Inform kernel about changed partition table in case parted couldn't
     pmb.chroot.root(["partprobe", "/dev/install"], check=False)
 
-    if not split:
+    if not split and not single_partition:
+        assert layout  # Initialized above for not single_partition case (mypy needs this)
         pmb.install.partitions_mount(device, layout, disk)
 
     pmb.install.format(args, layout, boot_label, root_label, disk)
@@ -933,7 +981,7 @@ def install_system_image(
     create_fstab(args, layout, chroot)
     if args.full_disk_encryption:
         logging.info("(native) create /etc/crypttab")
-        create_crypttab(args, layout, chroot)
+        create_crypttab(layout, chroot)
 
     # Run mkinitfs to pass UUIDs to cmdline
     logging.info(f"({chroot}) mkinitfs")
@@ -946,17 +994,37 @@ def install_system_image(
 
     # Just copy all the files
     logging.info(f"*** ({step + 1}/{steps}) FILL INSTALL BLOCKDEVICE ***")
-    copy_files_from_chroot(args, chroot)
+    copy_files_from_chroot(chroot, args.rsync, args.verbose)
     create_home_from_skel(args.filesystem, config.user)
-    configure_apk(args)
+    configure_apk(args.install_local_pkgs)
     copy_ssh_keys(config)
 
     # Don't try to embed firmware and cgpt on split images since there's no
     # place to put it and it will end up in /dev of the chroot instead
-    if not split:
+    if not split and not single_partition:
+        assert layout  # Initialized above for not single_partition case (mypy needs this)
         embed_firmware(args, chroot)
         write_cgpt_kpart(args, layout, chroot)
 
+    # Install GRUB to the PReP boot partition
+    if pmb.parse.deviceinfo().create_prep_boot and layout:
+        pmb.chroot.apk.install(["grub"], Chroot.native())
+        # GRUB is incapable of installing to /dev/installp1, it *needs* to install to
+        # /dev/loop0p1
+        pmb.chroot.root(
+            [
+                "grub-install",
+                "--target=powerpc-ieee1275",
+                "--directory=/mnt/install/usr/lib/grub/powerpc-ieee1275",
+                "--boot-directory=/mnt/install/boot",
+                # Since we are in a chroot, it cannot properly autodetect the modules,
+                # so we need to give it a hint
+                "--modules=part_gpt part_msdos disk normal linux boot search terminal",
+                "--no-nvram",
+                f"/dev/loop0p{layout['prep']}",
+            ]
+        )
+
     if disk:
         logging.info(f"Unmounting disk {disk} (this may take a while to sync, please wait)")
     pmb.chroot.shutdown(True)
@@ -996,7 +1064,9 @@ def install_system_image(
             pmb.chroot.user(["mv", "-f", sys_image_patched, sys_image], working_dir=workdir)
 
 
-def print_flash_info(device: str, deviceinfo: Deviceinfo, split: bool, have_disk: bool) -> None:
+def print_flash_info(
+    device: str, deviceinfo: Deviceinfo, split: bool, have_disk: bool, single_partition: bool
+) -> None:
     """Print flashing information, based on the deviceinfo data and the
     pmbootstrap arguments."""
     logging.info("")  # make the note stand out
@@ -1027,12 +1097,13 @@ def print_flash_info(device: str, deviceinfo: Deviceinfo, split: bool, have_disk
             logging.info(f"  {Chroot.native() / 'home/pmos/rootfs' / device}-root.img")
         else:
             logging.info(f"  {Chroot.native() / 'home/pmos/rootfs' / device}.img")
-            logging.info(
-                "  (NOTE: This file has a partition table, which"
-                " contains /boot and / subpartitions. That way we"
-                " don't need to change the partition layout on your"
-                " device.)"
-            )
+            if not single_partition:
+                logging.info(
+                    "  (NOTE: This file has a partition table, which"
+                    " contains /boot and / subpartitions. That way we"
+                    " don't need to change the partition layout on your"
+                    " device.)"
+                )
 
     # if current flasher supports vbmeta and partition is explicitly specified
     # in deviceinfo
@@ -1068,7 +1139,7 @@ def print_flash_info(device: str, deviceinfo: Deviceinfo, split: bool, have_disk
         logging.info(f"  {Chroot.rootfs(device) / 'boot/boot.img'}")
         logging.info("  (NOTE: This is not necessary if using a custom bootloader like U-Boot)")
 
-    if "boot" in flasher_actions:
+    if "boot" in flasher_actions or "boot_gki" in flasher_actions:
         logging.info(
             "  (NOTE: " + method + " also supports booting"
             " the kernel/initramfs directly without flashing."
@@ -1097,6 +1168,7 @@ def print_flash_info(device: str, deviceinfo: Deviceinfo, split: bool, have_disk
 def install_recovery_zip(args: PmbArgs, device: str, arch: Arch, steps: int) -> None:
     logging.info(f"*** ({steps}/{steps}) CREATING RECOVERY-FLASHABLE ZIP ***")
     chroot = Chroot(ChrootType.BUILDROOT, arch)
+    pmb.chroot.init(chroot)
     mount_device_rootfs(Chroot.rootfs(device), chroot)
     pmb.install.recovery.create_zip(args, chroot, device)
 
@@ -1185,6 +1257,7 @@ def install_on_device_installer(args: PmbArgs, step: int, steps: int) -> None:
         boot_label,
         "pmOS_install",
         args.split,
+        args.single_partition,
         args.disk,
     )
 
@@ -1332,9 +1405,6 @@ def create_device_rootfs(args: PmbArgs, step: int, steps: int) -> None:
         install_packages += context.config.extra_packages.split(",")
     if args.add:
         install_packages += args.add.split(",")
-    locale_is_set = config.locale != Config().locale
-    if locale_is_set:
-        install_packages += ["lang", "musl-locales"]
 
     pmaports_cfg = pmb.config.pmaports.read_config()
     # postmarketos-base supports a dummy package for blocking unl0kr install
@@ -1374,8 +1444,7 @@ def create_device_rootfs(args: PmbArgs, step: int, steps: int) -> None:
     # installed a hook without pmbootstrap - see #69 for more info)
     # Packages will be built if necessary as part of this step
     pmb.chroot.apk.install(install_packages, chroot)
-    flavor = pmb.chroot.other.kernel_flavor_installed(chroot)
-    pmb.chroot.initfs.build(flavor, chroot)
+    pmb.chroot.initfs.build(chroot)
 
     # Set the user password
     setup_login(args, config, chroot)
@@ -1387,8 +1456,7 @@ def create_device_rootfs(args: PmbArgs, step: int, steps: int) -> None:
     setup_timezone(chroot, config.timezone)
 
     # Set locale
-    if locale_is_set:
-        setup_locale(chroot, config.locale)
+    setup_locale(chroot, config.locale)
 
     # Set the hostname as the device name
     setup_hostname(device, config.hostname)
@@ -1413,6 +1481,13 @@ def install(args: PmbArgs) -> None:
     if args.on_device_installer:
         sanity_check_ondev_version(args)
 
+    # --single-partition implies --no-split. There is nothing to split if
+    # there is only a single partition.
+    if args.single_partition:
+        args.split = False
+        if deviceinfo.create_initfs_extra:
+            raise RuntimeError("--single-partition does not work for devices with initramfs-extra")
+
     # Number of steps for the different installation methods.
     if args.no_image:
         steps = 2
@@ -1446,10 +1521,23 @@ def install(args: PmbArgs) -> None:
         # Runs install_system_image twice
         install_on_device_installer(args, step, steps)
     else:
-        install_system_image(args, 0, chroot, step, steps, split=args.split, disk=args.disk)
+        install_system_image(
+            args,
+            0,
+            chroot,
+            step,
+            steps,
+            split=args.split,
+            disk=args.disk,
+            single_partition=args.single_partition,
+        )
 
     print_flash_info(
-        device, deviceinfo, args.split, True if args.disk and args.disk.is_absolute() else False
+        device,
+        deviceinfo,
+        args.split,
+        bool(args.disk and args.disk.is_absolute()),
+        args.single_partition,
     )
     print_sshd_info(args)
     print_firewall_info(args.no_firewall, deviceinfo.arch)
